import gzip
import hashlib
import logging
import mimetypes
import os
import re
import tempfile
from typing import IO

import aiohttp
import magic

from udata_hydra import config
from udata_hydra.utils import IOException

log = logging.getLogger("udata-hydra")


def compute_checksum_from_file(filename: str) -> str:
    """Compute sha1 in blocks"""
    sha1sum = hashlib.sha1()
    with open(filename, "rb") as f:
        block = f.read(2**16)
        while len(block) != 0:
            sha1sum.update(block)
            block = f.read(2**16)
    return sha1sum.hexdigest()


def extract_gzip(file_path: str) -> IO[bytes]:
    with gzip.open(file_path, "rb") as gz_file:
        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as temp_file:
            temp_file.write(gz_file.read())
    return temp_file


async def download_resource(
    url: str,
    headers: dict | None = None,
    max_size_allowed: int | None = None,
) -> tuple[IO[bytes], str]:
    """
    Attempts downloading a resource from a given url.
    Returns a tuple of (downloaded_file_object, detected_extension).
    Raises custom IOException if the resource is too large or if the URL is unreachable.
    """
    if (
        headers
        and max_size_allowed is not None
        and float(headers.get("content-length", -1)) > max_size_allowed
    ):
        raise IOException("File too large to download", url=url)

    tmp_file = tempfile.NamedTemporaryFile(
        dir=config.TEMPORARY_DOWNLOAD_FOLDER or None, delete=False
    )

    chunk_size = 1024
    i = 0
    too_large, download_error = False, None
    try:
        async with aiohttp.ClientSession(
            headers={"user-agent": config.USER_AGENT_FULL},
            raise_for_status=True,
        ) as session:
            async with session.get(url, allow_redirects=True) as response:
                async for chunk in response.content.iter_chunked(chunk_size):
                    if max_size_allowed is None or i * chunk_size < max_size_allowed:
                        tmp_file.write(chunk)
                    else:
                        too_large = True
                        break
                    i += 1
    except aiohttp.ClientResponseError as e:
        download_error = e
    finally:
        tmp_file.close()
        if too_large:
            os.remove(tmp_file.name)
            raise IOException("File too large to download", url=url)
        if download_error:
            os.remove(tmp_file.name)
            raise IOException("Error downloading CSV", url=url) from download_error

    detected_extension = ""

    if magic.from_file(tmp_file.name, mime=True) in [
        "application/x-gzip",
        "application/gzip",
    ]:
        # It's compressed - extract and determine extension from URL
        tmp_file = extract_gzip(tmp_file.name)

        # Extract any extension before .gz using regex
        match = re.search(r"\.([^.]+)\.gz$", url)
        if match:
            detected_extension = f".{match.group(1)}"
        else:
            detected_extension = ""
    else:
        # Not compressed - use magic to detect type
        mime_type = magic.from_file(tmp_file.name, mime=True)
        detected_extension = mimetypes.guess_extension(mime_type) or ""

    return tmp_file, detected_extension


async def download_file(url: str, fd):
    """Download a file from URL to a file descriptor"""
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            while True:
                chunk = await resp.content.read(1024)
                if not chunk:
                    break
                fd.write(chunk)


def remove_remainders(resource_id: str, extensions: list[str]) -> None:
    """Delete potential remainders from process that crashed"""
    for ext in extensions:
        if os.path.exists(f"{resource_id}.{ext}"):
            os.remove(f"{resource_id}.{ext}")



import importlib.metadata
import logging
import os
import re
import tomllib
from pathlib import Path

log = logging.getLogger("udata-hydra")


class Configurator:
    """Loads a dict of config from TOML file(s) and behaves like an object, ie config.VALUE"""

    configuration: dict = {}

    def __init__(self):
        if not self.configuration:
            self.configure()

    def configure(self) -> None:
        # load default settings
        with open(Path(__file__).parent / "config_default.toml", "rb") as f:
            configuration: dict = tomllib.load(f)

        # override with local settings
        local_settings = os.environ.get("HYDRA_SETTINGS", Path.cwd() / "config.toml")
        if Path(local_settings).exists():
            with open(Path(local_settings), "rb") as f:
                configuration.update(tomllib.load(f))

        self.configuration = configuration
        self.check()

        # add project metadata to config
        self.configuration["APP_NAME"] = "udata-hydra"
        self.configuration["APP_VERSION"] = importlib.metadata.version("udata-hydra")

    def override(self, **kwargs) -> None:
        self.configuration.update(kwargs)
        self.check()

    def check(self) -> None:
        """Sanity check on config"""
        assert self.MAX_POOL_SIZE >= self.BATCH_SIZE, "BATCH_SIZE cannot exceed MAX_POOL_SIZE"

    def __getattr__(self, __name):
        return self.configuration.get(__name)

    @property
    def __dict__(self):
        return self.configuration

    @property
    def USER_AGENT_FULL(self) -> str:
        """Build the complete user agent string with version"""
        if self.USER_AGENT and self.APP_VERSION:
            # Use regex to find pattern: / followed by version-like string
            pattern = r"/([^/\s]+)(?=\s|$)"
            result = re.sub(pattern, f"/{self.APP_VERSION}", self.USER_AGENT)
            # If no replacement was made (no version found), append it
            if result == self.USER_AGENT:
                return f"{self.USER_AGENT}/{self.APP_VERSION}"
            return result
        return "udata-hydra"


config = Configurator()



import asyncio
import logging
import time
from collections import defaultdict
from urllib.parse import urlparse

import aiohttp
from asyncpg import Record

from udata_hydra import config, context
from udata_hydra.crawl.helpers import (
    convert_headers,
    fix_surrogates,
    has_nice_head,
    is_domain_backoff,
)
from udata_hydra.crawl.preprocess_check_data import preprocess_check_data
from udata_hydra.db.resource import Resource
from udata_hydra.utils import queue

RESOURCE_RESPONSE_STATUSES = {
    "OK": "ok",
    "TIMEOUT": "timeout",
    "ERROR": "error",
    "BACKOFF": "backoff",
}

log = logging.getLogger("udata-hydra")


results: defaultdict = defaultdict(int)


async def check_batch_resources(to_parse: list[Record]) -> None:
    """Check a batch of resources"""
    context.monitor().set_status("Checking resources...")
    tasks: list = []
    async with aiohttp.ClientSession(
        timeout=None,
        headers={"user-agent": config.USER_AGENT_FULL},
    ) as session:
        for row in to_parse:
            tasks.append(
                check_resource(
                    url=row["url"],
                    resource=row,
                    session=session,
                    worker_priority="default" if row["priority"] else "low",
                )
            )
        for task in asyncio.as_completed(tasks):
            result = await task
            results[result] += 1
            context.monitor().refresh(results)


async def check_resource(
    url: str,
    resource: Record,
    session,
    sleep: float = 0,
    method: str = "head",
    worker_priority: str = "default",
    force_analysis: bool = False,
) -> str:
    log.debug(f"check {url}, sleep {sleep}, method {method}")

    # Import here to avoid circular import issues
    from udata_hydra.analysis.resource import analyse_resource

    if sleep:
        await asyncio.sleep(sleep)

    url_parsed = urlparse(url)
    domain = url_parsed.netloc

    if not domain:
        log.warning(f"[warning] not netloc in url, skipping {url}")
        # Process the check data. If it has changed, it will be sent to udata
        await preprocess_check_data(
            dataset_id=resource["dataset_id"],
            check_data={
                "resource_id": str(resource["resource_id"]),
                "url": url,
                "error": "Not netloc in url",
                "timeout": False,
            },
        )
        return RESOURCE_RESPONSE_STATUSES["ERROR"]

    should_backoff, reason = await is_domain_backoff(domain)
    if should_backoff:
        log.info(f"backoff {domain} ({reason})")
        # skip this URL, it will come back in a next batch
        await Resource.update(
            str(resource["resource_id"]), data={"status": "BACKOFF", "priority": False}
        )
        return RESOURCE_RESPONSE_STATUSES["BACKOFF"]

    try:
        start = time.time()
        timeout = aiohttp.ClientTimeout(total=5)
        _method = getattr(session, method)
        async with _method(url, timeout=timeout, allow_redirects=True) as resp:
            end = time.time()
            if method != "get" and not has_nice_head(resp):
                return await check_resource(
                    url,
                    resource,
                    session,
                    force_analysis=force_analysis,
                    method="get",
                    worker_priority=worker_priority,
                )
            resp.raise_for_status()

            # Preprocess the check data. If it has changed, it will be sent to udata
            new_check, last_check = await preprocess_check_data(
                dataset_id=resource["dataset_id"],
                check_data={
                    "resource_id": str(resource["resource_id"]),
                    "url": url,
                    "domain": domain,
                    "status": resp.status,
                    "headers": convert_headers(resp.headers),
                    "timeout": False,
                    "response_time": end - start,
                },
            )

            # Update resource status to TO_ANALYSE_RESOURCE
            await Resource.update(
                str(resource["resource_id"]), data={"status": "TO_ANALYSE_RESOURCE"}
            )

            # Enqueue the resource for analysis
            queue.enqueue(
                analyse_resource,
                check=new_check,
                last_check=last_check,
                force_analysis=force_analysis,
                worker_priority=worker_priority,
                _priority=worker_priority,
            )

            return RESOURCE_RESPONSE_STATUSES["OK"]

    except asyncio.exceptions.TimeoutError:
        # Process the check data. If it has changed, it will be sent to udata
        await preprocess_check_data(
            dataset_id=resource["dataset_id"],
            check_data={
                "resource_id": str(resource["resource_id"]),
                "url": url,
                "domain": domain,
                "timeout": True,
            },
        )

        # Reset resource status so that it's not forbidden to be checked again
        await Resource.update(str(resource["resource_id"]), data={"status": None})

        return RESOURCE_RESPONSE_STATUSES["TIMEOUT"]

    # TODO: debug AssertionError, should be caught in DB now
    # File "[...]aiohttp/connector.py", line 991, in _create_direct_connection
    # assert port is not None
    # UnicodeError: encoding with 'idna' codec failed (UnicodeError: label too long)
    # eg http://%20Localisation%20des%20acc%C3%A8s%20des%20offices%20de%20tourisme
    except (
        aiohttp.client_exceptions.ClientError,
        AssertionError,
        UnicodeError,
    ) as e:
        # if we get a 404, it might be that the resource's URL has changed since last catalog load
        # we compare the actual URL to the one we have here to handle these cases
        if getattr(e, "status", None) == 404 and config.UDATA_URI:
            handled = await handle_wrong_resource_url(
                resource,
                session,
                url,
                force_analysis,
                worker_priority,
            )
            if handled is not None:
                return handled

        error = getattr(e, "message", None) or str(e)
        # Process the check data. If it has changed, it will be sent to udata
        await preprocess_check_data(
            dataset_id=resource["dataset_id"],
            check_data={
                "resource_id": str(resource["resource_id"]),
                "url": url,
                "domain": domain,
                "timeout": False,
                "error": fix_surrogates(error),
                "headers": convert_headers(getattr(e, "headers", {})),
                "status": getattr(e, "status", None),
            },
        )

        log.warning(f"Crawling error for url {url}", exc_info=e)

        # Reset resource status so that it's not forbidden to be checked again
        await Resource.update(str(resource["resource_id"]), data={"status": None})

        return RESOURCE_RESPONSE_STATUSES["ERROR"]


async def handle_wrong_resource_url(
    resource: Record,
    session,
    url: str,
    force_analysis: bool,
    worker_priority: str,
):
    resource_id = resource["resource_id"]
    stable_resource_url = f"{config.UDATA_URI.replace('api/2', 'fr')}/datasets/r/{resource_id}"
    async with session.head(stable_resource_url) as resp:
        resp.raise_for_status()
        actual_url = resp.headers.get("location")
    if actual_url and url != actual_url:
        await Resource.update(resource_id, data={"url": actual_url})
        return await check_resource(
            actual_url,
            resource,
            session,
            force_analysis=force_analysis,
            method="head",
            worker_priority=worker_priority,
        )
    return



from .auth import token_auth_middleware
from .csv import detect_tabular_from_headers
from .errors import IOException, ParseException, handle_parse_exception
from .file import (
    compute_checksum_from_file,
    download_file,
    download_resource,
    extract_gzip,
    remove_remainders,
)
from .geojson import detect_geojson_from_headers_or_catalog
from .http import UdataPayload, get_request_params, send
from .queue import enqueue
from .timer import Timer

__all__ = [
    "token_auth_middleware",
    "detect_tabular_from_headers",
    "IOException",
    "ParseException",
    "handle_parse_exception",
    "compute_checksum_from_file",
    "download_file",
    "download_resource",
    "extract_gzip",
    "remove_remainders",
    "detect_geojson_from_headers_or_catalog",
    "UdataPayload",
    "get_request_params",
    "send",
    "enqueue",
    "Timer",
]



import json
import logging
import os
from datetime import datetime, timezone
from enum import Enum

import magic
from asyncpg import Record
from dateparser import parse as date_parser

from udata_hydra import config, context
from udata_hydra.analysis.csv import analyse_csv
from udata_hydra.analysis.geojson import analyse_geojson
from udata_hydra.crawl.calculate_next_check import calculate_next_check_date
from udata_hydra.db.check import Check
from udata_hydra.db.resource import Resource
from udata_hydra.db.resource_exception import ResourceException
from udata_hydra.utils import (
    IOException,
    Timer,
    UdataPayload,
    compute_checksum_from_file,
    detect_geojson_from_headers_or_catalog,
    detect_tabular_from_headers,
    download_resource,
    queue,
    send,
)


class Change(Enum):
    HAS_CHANGED = 1
    HAS_NOT_CHANGED = 2
    NO_GUESS = 3


log = logging.getLogger("udata-hydra")


async def analyse_resource(
    check: dict,
    last_check: dict | None,
    force_analysis: bool = False,
    worker_priority: str = "default",
) -> None:
    """
    Perform analysis on the resource designated by check_id:
    - change analysis
    - size (optional)
    - mime_type (optional)
    - checksum (optional)
    - launch csv_analysis if looks like a CSV response

    Will call udata if first check or changes found, and update check with optional infos
    """

    # Check if the resource is in the exceptions table
    exception: Record | None = await ResourceException.get_by_resource_id(str(check["resource_id"]))

    resource_id = check["resource_id"]
    dataset_id = check["dataset_id"]
    url = check["url"]
    headers = json.loads(check["headers"] or "{}")

    timer = Timer("analyse-resource", resource_id)

    log.debug(f"Analysis for resource {resource_id} in dataset {dataset_id}")

    # Update resource status to ANALYSING_RESOURCE
    resource: Record | None = await Resource.update(
        resource_id, data={"status": "ANALYSING_RESOURCE_HEAD"}
    )

    # let's see if we can infer a modification date on early hints based on harvest infos and headers
    change_status, change_payload = await detect_resource_change_on_early_hints(resource)

    # could it be a CSV or a GeoJSON? If we get hints, we will analyse the file further depending on change status
    is_tabular, file_format = detect_tabular_from_headers(check)
    is_geojson: bool = await detect_geojson_from_headers_or_catalog(check)
    if is_geojson:
        file_format = "geojson"
    max_size_allowed = None if exception else int(config.MAX_FILESIZE_ALLOWED[file_format])

    # if the change status is NO_GUESS or HAS_CHANGED, let's download the file to get more infos
    dl_analysis = {}
    tmp_file = None
    if change_status != Change.HAS_NOT_CHANGED or force_analysis:
        try:
            await Resource.update(resource_id, data={"status": "DOWNLOADING_RESOURCE"})
            tmp_file, _ = await download_resource(url, headers, max_size_allowed)
            timer.mark("download-resource")
        except IOException:
            dl_analysis["analysis:error"] = "File too large to download"
        else:
            await Resource.update(resource_id, data={"status": "ANALYSING_DOWNLOADED_RESOURCE"})
            # Get file size
            dl_analysis["analysis:content-length"] = os.path.getsize(tmp_file.name)
            # Get checksum
            dl_analysis["analysis:checksum"] = compute_checksum_from_file(tmp_file.name)
            # Check if checksum has been modified if we don't have other hints
            if change_status == Change.NO_GUESS:
                (
                    change_status,
                    change_payload,
                ) = await detect_resource_change_from_checksum(
                    new_checksum=dl_analysis["analysis:checksum"], last_check=last_check
                )
            dl_analysis["analysis:mime-type"] = magic.from_file(tmp_file.name, mime=True)
        finally:
            if tmp_file and not (is_tabular or is_geojson):
                os.remove(tmp_file.name)
            await Check.update(
                check["id"],
                {
                    "checksum": dl_analysis.get("analysis:checksum"),
                    "analysis_error": dl_analysis.get("analysis:error"),
                    "filesize": dl_analysis.get("analysis:content-length"),
                    "mime_type": dl_analysis.get("analysis:mime-type"),
                },
            )

    # if the file has not changed, update the current check with the infos of the previous one
    elif last_check:
        await Check.update(
            check["id"],
            {k: last_check.get(k) for k in ["checksum", "analysis_error", "filesize", "mime_type"]},
        )

    if change_status in (Change.HAS_CHANGED, Change.HAS_NOT_CHANGED):
        await update_check_with_modification_and_next_dates(
            change_analysis=change_payload or {},
            check_id=check["id"],
            last_check=last_check,
            has_changed=change_status == Change.HAS_CHANGED,
        )

    analysis_results = {"analysis:check_id": check["id"]} | dl_analysis | (change_payload or {})

    timer.stop()

    if change_status == Change.HAS_CHANGED or not last_check or force_analysis:
        if is_tabular and tmp_file:
            # Change status to TO_ANALYSE_CSV
            await Resource.update(resource_id, data={"status": "TO_ANALYSE_CSV"})
            # Analyse CSV and create a table in the CSV database
            queue.enqueue(
                analyse_csv,
                check=check,
                file_path=tmp_file.name,
                _priority="high" if worker_priority == "high" else "default",
                _exception=bool(exception),
            )
        elif is_geojson and tmp_file:
            await Resource.update(resource_id, data={"status": "TO_ANALYSE_GEOJSON"})
            queue.enqueue(
                analyse_geojson,
                check=check,
                file_path=tmp_file.name,
                _priority="high" if worker_priority == "high" else "default",
                _exception=bool(exception),
            )
        else:
            await Resource.update(resource_id, data={"status": None})

        # Send analysis result to udata
        queue.enqueue(
            send,
            dataset_id=dataset_id,
            resource_id=resource_id,
            document=UdataPayload(analysis_results),
            _priority="high",
        )

    else:
        await Resource.update(resource_id, data={"status": None})


async def update_check_with_modification_and_next_dates(
    change_analysis: dict, check_id: int, last_check: dict | None, has_changed: bool = False
) -> None:
    """Update check with last_modified date and next_check date only if resource has changed

    Args:
        change_analysis: dict with optional "analysis:last-modified-at" key
        check_id: the ID of the current check
        last_check: the last check data, if any
        has_changed: if the resource has changed since last check
    """
    last_modified: str | None = change_analysis.get("analysis:last-modified-at")
    if last_modified:
        last_modified_at: datetime = datetime.fromisoformat(last_modified)
        payload = {"detected_last_modified_at": last_modified_at}
        if has_changed:
            next_check_at: datetime = calculate_next_check_date(
                has_check_changed=True, last_check=last_check, last_modified_at=last_modified_at
            )
            payload["next_check_at"] = next_check_at
        await Check.update(check_id, payload)


async def detect_resource_change_from_checksum(
    new_checksum, last_check: dict | None
) -> tuple[Change, dict | None]:
    """
    Checks if resource checksum has changed over time
    Returns a tuple with a Change status and an optional payload:
    {
        "analysis:last-modified-at": last_modified_date,
        "analysis:last-modified-detection": "computed-checksum",
    }
    """
    if last_check and last_check.get("checksum"):
        if last_check.get("checksum") != new_checksum:
            return Change.HAS_CHANGED, {
                "analysis:last-modified-at": datetime.now(timezone.utc).isoformat(),
                "analysis:last-modified-detection": "computed-checksum",
            }
        elif last_check.get("detected_last_modified_at"):
            return Change.HAS_NOT_CHANGED, {
                "analysis:last-modified-at": last_check["detected_last_modified_at"].isoformat(),
                "analysis:last-modified-detection": "previous-check-detection",
            }
    # if the previous check did not have the info, we investigate further
    return Change.NO_GUESS, None


async def detect_resource_change_from_last_modified_header(
    data: dict,
) -> tuple[Change, dict | None]:
    # last modified header check

    if len(data) == 1 and data[0]["last_modified"]:
        last_modified_date = date_parser(data[0]["last_modified"])
        return Change.HAS_CHANGED, {
            "analysis:last-modified-at": last_modified_date.isoformat(),
            "analysis:last-modified-detection": "last-modified-header",
        }

    if len(data) == 1 or not data[0]["last_modified"]:
        return Change.NO_GUESS, None

    last_modified_date = date_parser(data[0]["last_modified"])
    payload = {
        "analysis:last-modified-at": last_modified_date.isoformat(),
        "analysis:last-modified-detection": "last-modified-header",
    }
    if data[0]["last_modified"] != data[1]["last_modified"]:
        return Change.HAS_CHANGED, payload
    return Change.HAS_NOT_CHANGED, payload


async def detect_resource_change_from_content_length_header(
    data: dict,
) -> tuple[Change, dict | None]:
    # content-length variation between current and last check
    if len(data) <= 1 or not data[0].get("content_length"):
        return Change.NO_GUESS, None
    if data[0].get("content_length") and data[1].get("content_length"):
        if data[0]["content_length"] != data[1]["content_length"]:
            return Change.HAS_CHANGED, {
                # if resource has changed, set last-modified to the current check's creation
                "analysis:last-modified-at": data[0]["created_at"].isoformat(),
                "analysis:last-modified-detection": "content-length-header",
            }
        # same content_length is not 100% certainly no change, but a good tradeoff to prevent many downloads
        elif data[1].get("detected_last_modified_at"):
            return Change.HAS_NOT_CHANGED, {
                # no change, using the last-modified from the previous check (passed on from check to check)
                "analysis:last-modified-at": data[1]["detected_last_modified_at"].isoformat(),
                "analysis:last-modified-detection": "previous-check-detection",
            }
    # if the previous check did not have the info, we investigate further
    return Change.NO_GUESS, None


async def detect_resource_change_on_early_hints(
    resource: Record | None,
) -> tuple[Change, dict | None]:
    """
    Try to guess if a resource has been modified from harvest and headers in check data:
    - last-modified header value if it can be found and parsed
    - content-length if it is found and changed over time (vs last checks)

    Returns a tuple with a Change status and an optional payload:
    {
        "analysis:last-modified-at": last_modified_date,
        "analysis:last-modified-detection": "detection-method",
    }
    """
    data = None

    if resource:
        # Fetch current and last check headers
        q = """
        SELECT
            created_at,
            headers->>'last-modified' as last_modified,
            headers->>'content-length' as content_length,
            detected_last_modified_at
        FROM checks
        WHERE resource_id = $1
        ORDER BY created_at DESC
        LIMIT 2
        """
        pool = await context.pool()
        async with pool.acquire() as connection:
            data = await connection.fetch(q, str(resource["resource_id"]))

    # not enough checks to make a comparison
    if not data:
        return Change.NO_GUESS, None

    # let's see if we can infer a modification date from harvest infos
    change_status, change_payload = await detect_resource_change_from_harvest(data, resource)
    if change_status != Change.NO_GUESS:
        return change_status, change_payload

    # if not, let's see if we can infer a modifification date from last-modified headers
    (
        change_status,
        change_payload,
    ) = await detect_resource_change_from_last_modified_header(data)
    if change_status != Change.NO_GUESS:
        return change_status, change_payload

    # if not, let's see if we can infer a modifification date from content-length header
    return await detect_resource_change_from_content_length_header(data)


async def detect_resource_change_from_harvest(
    checks_data: tuple, resource: Record | None
) -> tuple[Change, dict | None]:
    """
    Checks if resource has a harvest.modified_at
    Returns a tuple with a Change status and an optional payload:
    {
        "analysis:last-modified-at": last_modified_date,
        "analysis:last-modified-detection": "harvest-resource-metadata",
    }
    """
    if len(checks_data) == 1:
        return Change.NO_GUESS, None

    last_check = checks_data[1]

    if resource and resource.get("harvest_modified_at"):
        payload = {
            "analysis:last-modified-at": resource["harvest_modified_at"].isoformat(),
            "analysis:last-modified-detection": "harvest-resource-metadata",
        }
        if resource["harvest_modified_at"] == last_check["detected_last_modified_at"]:
            return Change.HAS_NOT_CHANGED, payload
        return Change.HAS_CHANGED, payload
    return Change.NO_GUESS, None
